/*  libGringotts - generic encryption-to-file library
 *  (c) 2002, Germano Rizzo <mano@pluto.linux.it>
 *
 *  libgringotts.h - general header file for libgringotts
 *  Author: Germano Rizzo
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
#ifndef LIBGRG_H
#define LIBGRG_H

#include <sys/types.h>

//defines the GRG_TOKEN encapsulation
//as the opaque structure grg_token
typedef struct grg_token *GRG_TOKEN;

//encryption algorythms
typedef enum
{
	GRG_RIJNDAEL_128 = 0x00,	//00000000
	GRG_AES = 0x00,		//alias for GRG_RIJNDAEL_128
	GRG_SERPENT = 0x10,	//00010000 (default)
	GRG_TWOFISH = 0x20,	//00100000
	GRG_CAST_256 = 0x30,	//00110000
	GRG_SAFERPLUS = 0x40,	//01000000
	GRG_LOKI97 = 0x50,	//01010000
	GRG_3DES = 0x60,	//01100000
	GRG_RIJNDAEL_256 = 0x70	//01110000
}
grg_crypt_algo;

//hashing algorythms
typedef enum
{
	GRG_SHA1 = 0x00,	//00000000
	GRG_RIPEMD_160 = 0x08	//00001000 (default)
}
grg_hash_algo;

//compression algorythm
typedef enum
{
	GRG_ZLIB = 0x00,	//00000000 (default)
	GRG_BZIP = 0x04		//00000100
}
grg_comp_algo;

//compression level
typedef enum
{
	GRG_LVL_NONE = 0x00,	//00000000
	GRG_LVL_FAST = 0x01,	//00000001
	GRG_LVL_GOOD = 0x02,	//00000010
	GRG_LVL_BEST = 0x03	//00000011 (default)
}
grg_comp_ratio;

//security level
typedef enum
{
	GRG_SEC_MIN,
	GRG_SEC_MED,
	GRG_SEC_MAX,
	GRG_SEC_PARANOIA
}
grg_security_lvl;

//masks
#define GRG_ENCRYPT_MASK		0x70	//01110000
#define GRG_HASH_MASK			0x08	//00001000
#define GRG_COMP_TYPE_MASK		0x04	//00000100
#define GRG_COMP_LVL_MASK		0x03	//00000011

//these enums can be &'ed to form the algorythm byte
//e.g.: default = 0x10 & 0x08 & 0x00 & 0x03 = 0x1b

//I/O Ok
#define GRG_OK							0

//I/O Errors
//error codes in writing
#define GRG_WRITE_COMP_ERR				-2
#define GRG_WRITE_ENC_INIT_ERR			-4
#define GRG_WRITE_FILE_ERR				-6
#define GRG_WRITE_TOO_BIG_ERR			-8

//error codes in reading
#define GRG_READ_FILE_ERR				-1
#define GRG_READ_MAGIC_ERR				-3
#define GRG_READ_CRC_ERR				-5
#define GRG_READ_PWD_ERR				-7
#define GRG_READ_ENC_INIT_ERR			-9
#define GRG_READ_TOO_BIG_ERR			-11
#define GRG_READ_UNSUPPORTED_VERSION	-13
#define GRG_READ_COMP_ERR				-15

typedef struct _grg_context *GRG_CTX;
typedef struct _grg_key *GRG_KEY;

//from libgrg_crypt.h

unsigned int	grg_get_key_size_static (const grg_crypt_algo crypt_algo);
unsigned int	grg_get_key_size (const GRG_CTX gctx);
unsigned int	grg_get_block_size_static (const grg_crypt_algo crypt_algo);
unsigned int	grg_get_block_size (const GRG_CTX gctx);
int		grg_validate_file (const GRG_CTX gctx, const unsigned char *path);
int		grg_update_gctx_from_file (GRG_CTX gctx, const unsigned char *path);
int		grg_decrypt_file (const GRG_CTX gctx, const GRG_KEY keystruct, const unsigned char *path, unsigned char **origData, unsigned long *origDim);
int		grg_encrypt_file (const GRG_CTX gctx, const GRG_KEY keystruct, const unsigned char *path, const unsigned char *origData, const unsigned long origDim);

//from libgrg_structs.h

//initializes a context from the given values
GRG_CTX	grg_context_initialize (const unsigned char *header, const grg_crypt_algo crypt_algo,
								const grg_hash_algo hash_algo, const grg_comp_algo comp_algo,
								const grg_comp_ratio comp_lvl, const grg_security_lvl sec_lvl);
//initializes a context from defaults
GRG_CTX	grg_context_initialize_defaults (const unsigned char *header);
//frees all the resources maintained by a context
void	grg_context_free (GRG_CTX gctx);

//get the various settings stored in a context
grg_crypt_algo		grg_ctx_get_crypt_algo (const GRG_CTX gctx);
grg_hash_algo		grg_ctx_get_hash_algo (const GRG_CTX gctx);
grg_comp_algo		grg_ctx_get_comp_algo (const GRG_CTX gctx);
grg_comp_ratio		grg_ctx_get_comp_ratio (const GRG_CTX gctx);
grg_security_lvl	grg_ctx_get_security_lvl (const GRG_CTX gctx);

//set the various settings stored in a context
void	grg_ctx_set_crypt_algo (GRG_CTX gctx, const grg_crypt_algo crypt_algo);
void	grg_ctx_set_hash_algo (GRG_CTX gctx, const grg_hash_algo hash_algo);
void	grg_ctx_set_comp_algo (GRG_CTX gctx, const grg_comp_algo comp_algo);
void	grg_ctx_set_comp_ratio (GRG_CTX gctx, const grg_comp_ratio comp_ratio);
void	grg_ctx_set_security_lvl (GRG_CTX gctx, const grg_security_lvl sec_level);

//GRG_KEY
GRG_KEY grg_key_gen (const unsigned char *pwd, const int pwd_len);
GRG_KEY grg_key_clone (const GRG_KEY src);
int grg_key_compare (const GRG_KEY k1, const GRG_KEY k2);
void grg_key_free (const GRG_CTX gctx, GRG_KEY key);

//from libgrg_utils.h
unsigned char	*grg_rnd_seq (const GRG_CTX gctx, const unsigned int size);
unsigned char	grg_rnd_chr (const GRG_CTX gctx);
void	grg_free (const GRG_CTX gctx, void *alloc_data, const long dim);
double	grg_ascii_pwd_quality (const unsigned char *pwd, const unsigned long pwd_len);
double	grg_file_pwd_quality (const unsigned char *pwd_path);

#endif
