/*  Gringotts - a small utility to safe-keep sensitive data
 *  (c) 2002, Germano Rizzo <mano@pluto.linux.it>
 *
 *  grg_prefs_io.c - preferences load/saving
 *  Author: Germano Rizzo
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/stat.h>

#include <gtk/gtk.h>

#include "../libgringotts/libgringotts.h"

#include "grg_prefs.h"
#include "grg_prefs_io.h"
#include "grg_defs.h"
#include "gringotts.h"

gint
grg_save_prefs (void)
{
	guchar *path, *row, algo, *grg_pref_file_local;
	gint fd;

	path = g_build_filename (g_get_home_dir (), ".gringotts.conf", NULL);
	fd = open (path, O_WRONLY | O_CREAT | O_TRUNC, S_IRUSR | S_IRGRP | S_IROTH | S_IWUSR);
	g_free (path);

	if (fd < 3)
	{
		close (fd);
		return GRG_PREFS_IO_ERROR;
	}

	algo = (guchar) (grg_ctx_get_crypt_algo(gctx) | grg_ctx_get_hash_algo(gctx) | grg_ctx_get_comp_algo(gctx) |
			 grg_ctx_get_comp_ratio(gctx));

	//saves the algorythm
	row = g_strdup_printf ("<algo_code>\n%02x\n</algo_code>\n", algo);
	write (fd, row, strlen (row));
	g_free (row);

	//saves the startup file
	grg_pref_file_local = get_pref_file ();
	if(grg_pref_file_local)
	{
		row = g_strdup_printf ("<startup_file>\n%s\n</startup_file>\n", grg_pref_file_local);
		g_free (grg_pref_file_local);
		write (fd, row, strlen (row));
		g_free (row);
	}

	//saves the backup files preference
	row = g_strdup_printf ("<bak_files>\n%c\n</bak_files>\n", grg_prefs_bak_files ? '1' : '0');
	write (fd, row, strlen (row));
	g_free (row);

	//saves the file overwriting warning preference
	row = g_strdup_printf ("<overwrite_warn>\n%c\n</overwrite_warn>\n", grg_prefs_warn4overwrite ? '1' : '0');
	write (fd, row, strlen (row));
	g_free (row);

	//saves the splash screen preference
	row = g_strdup_printf ("<display_splash_screen>\n%c\n</display_splash_screen>\n", grg_prefs_splash ? '1' : '0');
	write (fd, row, strlen (row));
	g_free (row);

	close (fd);

	return GRG_OK;
}

#define IGNORE		'\0'
#define ALGO		'a'
#define START_FILE	's'
#define OVER		'o'
#define BACKUP		'b'
#define SPLASH		'd'

static void
introduce_pref (GMarkupParseContext *context,
				const gchar	*element_name,
				const gchar	**attribute_names,
				const gchar	**attribute_values,
				gpointer	user_data,
				GError	**error)
{
	*((gchar *) user_data) = element_name[0];
}

static void
collect_pref (GMarkupParseContext *context,
				const gchar *text,
				gsize text_len,  
				gpointer user_data,
				GError **error)
{
	switch (*((gchar *) user_data))
	{
		case ALGO:
		{
			gchar algo1=0, algo2=0, algo;
			gint i,pos=0;
			for (i=0; i<text_len; i++) //strips whitespaces
				if((text[i] != ' ') && (text[i] != '\t') && (text[i] != '\n')){
					if(!pos) //pos=0
						algo1 = text[i];
					else
						algo2 = text[i];
					pos++;
			}
	
			if (!(((algo1 >= '0') && (algo1 <= '9')) || ((algo1 >= 'a') && (algo1 <= 'f')))
				|| !(((algo2 >= '0') && (algo2 <= '9')) || ((algo2 >= 'a') && (algo2 <= 'f'))))
				break;
			//form the algo byte and setup values
			if (algo1 < 'a')
				algo1 -= '0';
			else
				algo1 -= 'W';	//'a'-10
		
			if (algo2 < 'a')
				algo2 -= '0';
			else
				algo2 -= 'W';	//'a'-10
		
			algo = 0;
			algo |= (guchar) ((algo1 << 4) & 0xf0);
			algo |= (guchar) (algo2 & 0x0f);
		
			grg_ctx_set_crypt_algo(gctx, (grg_crypt_algo) (algo & GRG_ENCRYPT_MASK));
			grg_ctx_set_hash_algo(gctx, (grg_hash_algo) (algo & GRG_HASH_MASK));
			grg_ctx_set_comp_algo(gctx, (grg_comp_algo) (algo & GRG_COMP_TYPE_MASK));
			grg_ctx_set_comp_ratio(gctx, (grg_comp_ratio) (algo & GRG_COMP_LVL_MASK));

			break;
		}
		case START_FILE:
		{
			if(text_len == 0)
				set_pref_file(NULL);
			else
			{
				gchar *file = g_strstrip(g_strndup(text, text_len));
				gint fdt = open (file, O_RDONLY);
				if (fdt < 3)
					set_pref_file(NULL);
				else
					set_pref_file (file);
				close (fdt);
				g_free (file);
			}
			break;
		}
		case OVER:
		{
			int i=0;
			while(text[i] == 10 || text[i] == ' ' || text[i] == '\t' || text[i] == 13)
				i++;
			grg_prefs_warn4overwrite = (text[i] == '0' ? FALSE : TRUE);
			break;
		}
		case BACKUP:
		{
			int i=0;
			while(text[i] == 10 || text[i] == ' ' || text[i] == '\t' || text[i] == 13)
				i++;
			grg_prefs_bak_files = (text[i] == '0' ? FALSE : TRUE);
			break;
		}
		case SPLASH:
		{
			int i=0;
			while(text[i] == 10 || text[i] == ' ' || text[i] == '\t' || text[i] == 13)
				i++;
			grg_prefs_splash = (text[i] == '0' ? FALSE : TRUE);
			break;
		}
		case IGNORE:
			break;
		default:
	}
	*((gchar *) user_data) = IGNORE;
}

gint
grg_load_prefs (void)
{
	guchar *path, *content, active_opt=0;
	gint fd, end;
	GMarkupParser *context = (GMarkupParser *) g_malloc (sizeof (GMarkupParser));
	GMarkupParseContext *parser;
	GError *err=NULL;

	path = g_build_filename (g_get_home_dir (), ".gringotts.conf", NULL);
	fd = open (path, O_RDONLY);
	g_free (path);

	if (fd < 3)
	{
		close (fd);
		return GRG_PREFS_IO_ERROR;
	}
	
	end = lseek (fd, 0, SEEK_END);
	lseek (fd, 0, SEEK_SET);

	context->start_element = introduce_pref;
	context->end_element = NULL;
	context->text = collect_pref;
	context->passthrough = NULL;
	context->error = NULL;

	parser = g_markup_parse_context_new (context, 0, &active_opt, NULL);

	content = (gchar *) mmap (NULL, end, PROT_READ, MAP_PRIVATE, fd, 0);

	g_markup_parse_context_parse (parser, content, end, &err);
	if(!err)
		g_markup_parse_context_end_parse (parser, &err);

	g_markup_parse_context_free (parser);
	if(err)
		g_error_free (err);
	g_free (context);
	
	munmap (content, end);
	close (fd);

	return GRG_OK;
}
