/*  Gringotts - a small utility to safe-keep sensitive data
 *  (c) 2002, Germano Rizzo <mano@pluto.linux.it>
 *
 *  grg_pwd.c - widgets used to get passwords
 *  Author: Germano Rizzo
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/mman.h>

#include <gtk/gtk.h>

#include "../libgringotts/libgringotts.h"

#include "grg_pwd.h"
#include "grg_defs.h"
#include "grg_widgets.h"
#include "gringotts.h"

#define TYPE_PWD	0
#define TYPE_FILE	1

#define NEW_RADIO_BUTTON(widget, list, call, value, label, box); \
	widget = gtk_radio_button_new_with_label(list, label); \
	g_signal_connect (G_OBJECT (widget), "toggled", G_CALLBACK (call), GINT_TO_POINTER(value)); \
	gtk_box_pack_start(GTK_BOX(box), widget, FALSE, FALSE, GRG_PAD);

#define NEW_LABEL(box, text); \
	{ \
		GtkWidget *new_lbl; \
		new_lbl = gtk_label_new(text); \
		gtk_misc_set_alignment(GTK_MISC(new_lbl), 0, 0); \
		gtk_box_pack_start(GTK_BOX(box), new_lbl, FALSE, FALSE, GRG_PAD); \
	}

#define SWAP_BUTTONS(widget, stock, block, unblock); \
	gtk_button_set_label (GTK_BUTTON(widget), stock); \
	g_signal_handler_block (widget, block); \
	g_signal_handler_unblock (widget, unblock);

#define NEW_SEPARATOR(box); \
	gtk_box_pack_start (GTK_BOX (box), gtk_hseparator_new (), FALSE, FALSE, GRG_PAD);

static void
meta_browse (GtkWidget * data, GtkWidget * entry)
{
	GtkWidget *file_selector;
	GtkWidget *dlg =
		GTK_WIDGET (GTK_WIDGET ((GTK_WIDGET (data->parent))->parent)->
			    parent);
	gint res;

	file_selector = gtk_file_selection_new (_("Open..."));
	gtk_window_set_transient_for (GTK_WINDOW (file_selector),
				      GTK_WINDOW (dlg));
	gtk_widget_show (file_selector);
	res = gtk_dialog_run (GTK_DIALOG (file_selector));
	if (res == GTK_RESPONSE_OK)
		gtk_entry_set_text (GTK_ENTRY (entry),
				    gtk_file_selection_get_filename
				    (GTK_FILE_SELECTION (file_selector)));
	gtk_widget_destroy (file_selector);
}

static GRG_KEY
read_pwd_file (const gchar *path, GtkWidget * dlg)
{
	gint fd, len;
	gchar *pwd;
	struct stat buf;
	GRG_KEY key;
	
	fd = open (path, O_RDONLY);

	if (fd < 0)
	{
		grg_msg (_("The file does not exist"),
			 GTK_MESSAGE_ERROR, dlg);
		return NULL;
	}
	
	if (fstat (fd, &buf) < 0)
	{
		grg_msg (_("The file cannot be opened"),
			 GTK_MESSAGE_ERROR, dlg);
		close (fd);
		return NULL;
	}
	
	if (!S_ISREG(buf.st_mode))
	{
		grg_msg (_("You must specify a regular file"),
			 GTK_MESSAGE_ERROR, dlg);
		close (fd);
		return NULL;
	}

	len = buf.st_size;

	pwd = mmap (NULL, len, PROT_READ, MAP_PRIVATE, fd, 0);
	key = grg_key_gen (pwd, len);
	munmap (pwd, len);
	close (fd);
	return key;
}

/*******************
* CHANGE PWD DIALOG *
 *******************/
static GtkWidget *label, *hbox_file, *vbox_pwd;
static GtkWidget *file_entry, *question, *question2, *quality;

static void
toggle_pwd_chg_file (GtkWidget * data, gpointer value)
{
	if (!gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (data)))
		return;

	if (GPOINTER_TO_INT (value) == TYPE_PWD)
	{
		gtk_label_set_text (GTK_LABEL (label),
				    _("Enter new password:"));
		gtk_entry_set_text (GTK_ENTRY (question), "");
		gtk_entry_set_text (GTK_ENTRY (question2), "");
		gtk_widget_hide (hbox_file);
		gtk_widget_show_all (vbox_pwd);
	}
	else
	{
		gtk_label_set_text (GTK_LABEL (label), _("Choose file:"));
		gtk_entry_set_text (GTK_ENTRY (file_entry), "");
		gtk_widget_hide (vbox_pwd);
		gtk_widget_show_all (hbox_file);
	}

	return;
}

static void
vis_quality (gpointer ignore, gpointer type)
{
	if (GPOINTER_TO_INT (type) == TYPE_PWD)
	{
		gint bout;
		guchar *sq = g_locale_from_utf8 (gtk_entry_get_text (GTK_ENTRY (question)), -1, NULL, &bout, NULL);

		gtk_progress_bar_set_fraction (GTK_PROGRESS_BAR (quality), grg_ascii_pwd_quality (sq, bout));

		grg_free (gctx, sq, bout);
	}
	else
		gtk_progress_bar_set_fraction (GTK_PROGRESS_BAR (quality),
			grg_file_pwd_quality ((guchar *) gtk_entry_get_text (GTK_ENTRY (file_entry))));
}

/**
 * grg_new_pwd_dialog:
 * @parent: the parent of the dialog to create.
 *
 * Asks for a new password, validating it (>4 chars).
 *
 * Returns: the new password, or NULL if user Cancel-ed
 */
GRG_KEY
grg_new_pwd_dialog (GtkWidget * parent)
{
	GtkWidget *dialog, *label2;
	GtkWidget *chk_file, *chk_pwd;
	GtkWidget *browse;
	GRG_KEY key = NULL;

	dialog = gtk_dialog_new_with_buttons (_("New password"), GTK_WINDOW (parent), GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
						  GTK_STOCK_OK, GTK_RESPONSE_OK, GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL, NULL);

	gtk_dialog_set_default_response (GTK_DIALOG (dialog),
					 GTK_RESPONSE_OK);
	gtk_container_set_border_width (GTK_CONTAINER (dialog), 3);
	gtk_box_set_spacing (GTK_BOX (GTK_DIALOG (dialog)->vbox), 3);
	gtk_window_set_resizable (GTK_WINDOW (dialog), FALSE);

	NEW_LABEL (GTK_DIALOG (dialog)->vbox, _("Choose password type:"));

	NEW_RADIO_BUTTON (chk_pwd, NULL, toggle_pwd_chg_file, TYPE_PWD,
			  _("String"), GTK_DIALOG (dialog)->vbox);
	NEW_RADIO_BUTTON (chk_file, gtk_radio_button_get_group(GTK_RADIO_BUTTON(chk_pwd)),
			  toggle_pwd_chg_file, TYPE_FILE, _("File"), GTK_DIALOG (dialog)->vbox);
	
	NEW_SEPARATOR (GTK_DIALOG (dialog)->vbox);

	label = gtk_label_new (_("Enter new password:"));
	gtk_box_pack_start_defaults (GTK_BOX (GTK_DIALOG (dialog)->vbox),
				     label);

	vbox_pwd = gtk_vbox_new (FALSE, GRG_PAD);
	gtk_box_pack_start_defaults (GTK_BOX (GTK_DIALOG (dialog)->vbox),
				     vbox_pwd);

	label2 = gtk_label_new (_("Enter it again for confirmation:"));

	question = gtk_entry_new ();
	gtk_entry_set_visibility (GTK_ENTRY (question), FALSE);
	g_signal_connect (G_OBJECT (question), "changed",
			  G_CALLBACK (vis_quality),
			  GINT_TO_POINTER (TYPE_PWD));

	question2 = gtk_entry_new ();
	gtk_entry_set_visibility (GTK_ENTRY (question2), FALSE);

	gtk_box_pack_start_defaults (GTK_BOX (vbox_pwd), question);
	gtk_box_pack_start_defaults (GTK_BOX (vbox_pwd), label2);
	gtk_box_pack_start_defaults (GTK_BOX (vbox_pwd), question2);

	hbox_file = gtk_hbox_new (FALSE, GRG_PAD);

	file_entry = gtk_entry_new ();
	g_signal_connect (G_OBJECT (file_entry), "changed",
			  G_CALLBACK (vis_quality),
			  GINT_TO_POINTER (TYPE_FILE));
	browse = gtk_button_new_from_stock (GTK_STOCK_OPEN);
	g_signal_connect (G_OBJECT (browse), "clicked",
			  G_CALLBACK (meta_browse), (gpointer) file_entry);
	gtk_box_pack_start_defaults (GTK_BOX (hbox_file), file_entry);
	gtk_box_pack_start_defaults (GTK_BOX (hbox_file), browse);

	NEW_SEPARATOR (GTK_DIALOG (dialog)->vbox);

	quality = gtk_progress_bar_new ();
	gtk_progress_bar_set_text (GTK_PROGRESS_BAR (quality),
				   _("password quality"));
	gtk_progress_bar_set_fraction (GTK_PROGRESS_BAR (quality), 0.0);

	gtk_box_pack_start_defaults (GTK_BOX (GTK_DIALOG (dialog)->vbox),
				     quality);

	//show all...
	gtk_widget_grab_focus (question);
	gtk_widget_show_all (GTK_DIALOG (dialog)->vbox);
	//...but the file selection part
	gtk_box_pack_start_defaults (GTK_BOX (GTK_DIALOG (dialog)->vbox),
				     hbox_file);
	gtk_box_reorder_child (GTK_BOX (GTK_DIALOG (dialog)->vbox), hbox_file,
			       7);

	while (TRUE)
	{
		gint res = gtk_dialog_run (GTK_DIALOG (dialog));
		grg_key_free (gctx, key);
		key = NULL;

		if (res == GTK_RESPONSE_CANCEL)
			break;

		if (gtk_toggle_button_get_active
		    (GTK_TOGGLE_BUTTON (chk_pwd)))
		{		//the password is a string
			gchar *ret1 = (gchar *) gtk_entry_get_text (GTK_ENTRY (question));
			gchar *ret2 = (gchar *) gtk_entry_get_text (GTK_ENTRY (question2));
			gint pwd_len = strlen (ret1);

			if (pwd_len < 4)
			{
				grg_msg (_
					 ("The password is too short, it must be at least 4 chars"),
					 GTK_MESSAGE_ERROR, dialog);

				gtk_entry_set_text (GTK_ENTRY (question), "");
				gtk_entry_set_text (GTK_ENTRY (question2), "");

				continue;
			}

			if (strncmp (ret1, ret2, pwd_len))
			{
				grg_msg (_("The two passwords are different"),
					 GTK_MESSAGE_ERROR, dialog);

				gtk_entry_set_text (GTK_ENTRY (question), "");
				gtk_entry_set_text (GTK_ENTRY (question2), "");

				continue;
			}
			
			key = grg_key_gen (ret1, pwd_len);

			break;
		}
		else
		{		//the password is a file
			const guchar *path = gtk_entry_get_text (GTK_ENTRY (file_entry));

			key = read_pwd_file (path, dialog);
			if (key)
				break;

			continue;
		}
	}

	gtk_widget_destroy (dialog);
	
	return key;
}

/****************
* ASK PWD DIALOG *
 ****************/
static GtkWidget *util_button, *entry;
static guint sigclear, sigbrowse;

static void
clear_entry (GtkWidget * data, GtkWidget * dentry)
{
	gtk_entry_set_text (GTK_ENTRY (dentry), "");
}

static void
toggle_pwd_file (GtkWidget * data, gpointer value)
{
	if (!gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (data)))
		return;

	if (GPOINTER_TO_INT (value) == TYPE_PWD)
	{
		SWAP_BUTTONS (util_button, GTK_STOCK_CLEAR, sigbrowse,
			      sigclear);
		gtk_entry_set_text (GTK_ENTRY (entry), "");
		gtk_entry_set_visibility (GTK_ENTRY (entry), FALSE);
	}
	else
	{
		SWAP_BUTTONS (util_button, GTK_STOCK_OPEN, sigclear,
			      sigbrowse);
		gtk_entry_set_text (GTK_ENTRY (entry), "");
		gtk_entry_set_visibility (GTK_ENTRY (entry), TRUE);
	}
}

GRG_KEY
grg_ask_pwd_dialog (GtkWidget * parent)
{
	gint response;
	GtkWidget *dlg;
	GtkWidget *chk_file, *chk_pwd, *hbox;
	GRG_KEY key = NULL;

	dlg = gtk_dialog_new_with_buttons (_("Enter password"),
					   GTK_WINDOW (parent),
					   GTK_DIALOG_MODAL |
					   GTK_DIALOG_DESTROY_WITH_PARENT,
					   GTK_STOCK_OK, GTK_RESPONSE_OK,
					   GTK_STOCK_CANCEL,
					   GTK_RESPONSE_CANCEL, NULL);
	gtk_dialog_set_default_response (GTK_DIALOG (dlg), GTK_RESPONSE_CANCEL);
	gtk_container_set_border_width (GTK_CONTAINER (dlg), 3);
	gtk_box_set_spacing (GTK_BOX (GTK_DIALOG (dlg)->vbox), 3);
	gtk_window_set_resizable (GTK_WINDOW (dlg), FALSE);

	NEW_LABEL (GTK_DIALOG (dlg)->vbox, _("Choose password type:"));
	NEW_RADIO_BUTTON (chk_pwd, NULL, toggle_pwd_file, TYPE_PWD,
			  _("String"), GTK_DIALOG (dlg)->vbox);
	NEW_RADIO_BUTTON (chk_file, gtk_radio_button_get_group(GTK_RADIO_BUTTON(chk_pwd)),
			  toggle_pwd_file, TYPE_FILE, _("File"), GTK_DIALOG (dlg)->vbox);
	NEW_SEPARATOR (GTK_DIALOG (dlg)->vbox);

	hbox = gtk_hbox_new (FALSE, GRG_PAD);
	gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dlg)->vbox), hbox, FALSE,
				FALSE, 0);
	entry = gtk_entry_new ();
	gtk_entry_set_visibility (GTK_ENTRY (entry), FALSE);
	gtk_box_pack_start (GTK_BOX (hbox), entry, FALSE, FALSE, 0);
	util_button = gtk_button_new_from_stock (GTK_STOCK_CLEAR);
	gtk_button_set_use_stock (GTK_BUTTON (util_button), TRUE);
	sigclear =
		g_signal_connect (G_OBJECT (util_button), "clicked",
				  G_CALLBACK (clear_entry), (gpointer) entry);
	sigbrowse =
		g_signal_connect (G_OBJECT (util_button), "clicked",
				  G_CALLBACK (meta_browse), (gpointer) entry);
	g_signal_handler_block (util_button, sigbrowse);
	gtk_box_pack_start (GTK_BOX (hbox), util_button, FALSE, FALSE, 0);

	gtk_widget_grab_focus (entry);
	gtk_widget_show_all (dlg);

	while (TRUE)
	{
		clear_entry (NULL, entry);
		response = gtk_dialog_run (GTK_DIALOG (dlg));
		grg_key_free (gctx, key);
		key = NULL;

		if (response == GTK_RESPONSE_CANCEL)
			break;

		if (gtk_toggle_button_get_active
		    (GTK_TOGGLE_BUTTON (chk_pwd)))
		{
			key = grg_key_gen (gtk_entry_get_text (GTK_ENTRY (entry)), strlen (gtk_entry_get_text (GTK_ENTRY (entry))));
			break;
		}
		else
		{
			const guchar *path =
				gtk_entry_get_text (GTK_ENTRY (entry));
			key = read_pwd_file (path, dlg);

			if (key)
				break;
		}
	}

	gtk_widget_destroy (GTK_WIDGET (dlg));

	return key;
}
