/*  Gringotts - a small utility to safe-keep sensitive data
 *  (c) 2002, Germano Rizzo <mano@pluto.linux.it>
 *
 *  grg_safe.c - security management for the program
 *  Author: Germano Rizzo
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#define __USE_GNU		//to use the global variable <environ> in stdlib.h

#include <gtk/gtk.h>
#include <gdk-pixbuf/gdk-pixbuf.h>
#include <mcrypt.h>

#include "gringotts.h"
#include "grg_defs.h"
#include "grg_safe.h"
#include "grg_widgets.h"

#include "../libgringotts/libgringotts.h"

#include <stdlib.h>
#include <string.h>
#include <regex.h>
#include <unistd.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>
#ifdef HAVE_SYS_FSUID_H
#include <sys/fsuid.h>
#endif
#include <sys/time.h>
#include <sys/types.h>
#include <sys/resource.h>

#define GRG_SAFE			0
#define GRG_SLIGHTLY_UNSAFE	1
#define GRG_UNSAFE			2

static gboolean mem_safe = FALSE, ptrace_safe = FALSE;
static gint safety_level = GRG_SAFE;

gboolean
grg_security_filter (gboolean rootCheck)
{
	gint canary;
	struct rlimit *rl;
#ifdef ENV_CHECK
	regex_t regexp1, regexp2;
	gchar *lang, *display, *xauth;
#endif

	//drop eventual group root privileges
	setgid(getgid ());
	setgid(getgid ()); //twice for counter "saved IDs", cfr. Secure Programming HowTo
#ifdef HAVE_SYS_FSUID_H
	setfsgid(getgid ());
	setfsgid(getgid ());
#endif

	if (!geteuid())
	//the process is SUID root. I can mlockall() the memory in order to avoid swapping.
	{
#ifdef HAVE_MLOCK
		gint res = mlockall (MCL_FUTURE);

		if (res)
			g_critical ("%s", _("The process is setuid root, but I can't lock memory paging"));
		else
			mem_safe = TRUE;
#endif

		//drop root privileges
		setuid(getuid());
		setuid(getuid());
#ifdef HAVE_SYS_FSUID_H
		setfsuid(getuid());
		setfsuid(getuid());
#endif

#ifdef HAVE_SYS_FSUID_H
		if(getuid () && (!setuid(0) || !setfsuid(0)))
#else
		if(getuid () && !setuid(0))
#endif
		{
			g_critical ("%s", _("Cannot drop root privileges. Someone is hacking this process. I cannot go on"));
			return FALSE;
		}
		ptrace_safe = TRUE;
	}

#ifdef HAVE_SYS_FSUID_H
	if(getuid () && (!setgid(0) || !setfsgid(0)))
#else
	if(getuid () && !setgid(0))
#endif
	{
		g_critical ("%s", _("Cannot drop root privileges. Someone is hacking this process. I cannot go on"));
		return FALSE;
	}

	if (!rootCheck && (!getuid () || !geteuid () || !getgid () || !getegid ()))
	//forbid root usage
	{
		g_critical ("%s", _("For security reasons, you cannot run Gringotts as root. Try using --disable-root-filter in ./configure"));
		return FALSE;
	}

	//set and check core dump generation
	rl = (struct rlimit *) malloc(sizeof(struct rlimit));
	rl->rlim_cur = rl->rlim_max = 0;
	setrlimit(RLIMIT_CORE, rl);
	getrlimit(RLIMIT_CORE, rl);
	if(rl->rlim_cur || rl->rlim_max) //no need to give any message, it should be impossible
		return FALSE;
	g_free(rl);

	//checks that stderr, stdin & stdout are opened
	canary = open("/dev/null", O_CREAT);
	if ((canary >= 0) && (canary <= 2))
	{
		g_critical ("%s", _("stdin, stdout and/or stderr are invalid. Cannot continue."));
		close (canary);
		return FALSE;
	}
	close (canary);

	//extract needed environmental vars, validate, erase environment,
	//and re-set them (see Secure Programming HowTo, sect.4.2)

#ifdef ENV_CHECK
	//extract
	lang = getenv ("LANG");
	display = getenv ("DISPLAY");
	xauth = getenv ("XAUTHORITY");
	//validate
	regcomp (&regexp1, "[[:alpha:]][[:alnum:]_,+@\\-\\.=]*",
		 REG_EXTENDED);
	regcomp (&regexp2,
		 "[[:digit:]{1,3}\\.[:digit:]{1,3}\\.[:digit:]{1,3}\\.[:digit:]{1,3}]?:[[:digit:]\\.]+",
		 REG_EXTENDED);
	if (((lang != NULL) && (regexec (&regexp1, lang, 0, NULL, 0)))
	    || ((xauth != NULL) && (!g_file_test (xauth, G_FILE_TEST_EXISTS)))
	    || (regexec (&regexp2, display, 0, NULL, 0)))
	{
		g_critical ("%s", _("Invalid environment variables. They could have been manipulated."));
		return FALSE;
	}
	//erase
#ifdef HAVE_CLEARENV
	clearenv();
#else
#ifdef HAVE_ENVIRON
	{
		extern char **environ;

		environ = NULL;
	}
#endif
#endif

	//re-set (warning: don't free() the g_strconcat'ed strings)
	if (lang != NULL)
		putenv (g_strconcat ("LANG=", lang, NULL));
	putenv (g_strconcat ("DISPLAY=", display, NULL));
	if (xauth != NULL)
		putenv (g_strconcat ("XAUTHORITY=", xauth, NULL));
	putenv (g_strconcat ("HOME=", g_get_home_dir (), NULL));
#endif

	//this enables a stronger check on malloc() routines
#ifdef MAINTAINER_MODE
	putenv ("MALLOC_CHECK_=2");
#else
	putenv ("MALLOC_CHECK_=0");
#endif

	//initializes the security level indicator
	if (!(geteuid () && getegid () && getuid () && getgid ()))
		safety_level = GRG_UNSAFE;

#ifdef HAVE_MLOCK
	if(!mem_safe)
		safety_level = GRG_UNSAFE;
#endif

	if(!ptrace_safe)
		safety_level = GRG_UNSAFE;

#ifndef ENV_CHECK
	safety_level = GRG_UNSAFE;
#endif

#ifndef STRONG_RANDOM
	if (safety_level == GRG_SAFE)
		safety_level = GRG_SLIGHTLY_UNSAFE;
#endif

#ifndef ROOT_FILTER
	if (safety_level == GRG_SAFE)
		safety_level = GRG_SLIGHTLY_UNSAFE;
#endif

	return TRUE;
}

const gchar *red_xpm[], *yellow_xpm[], *green_xpm[], *optimal_xpm[];

GtkWidget
*grg_get_security_button (void)
{
	GdkPixbuf *pix;
	GtkWidget *img;

	switch (safety_level){
		case GRG_SAFE:
			pix = gdk_pixbuf_new_from_xpm_data (optimal_xpm);
			break;
		case GRG_SLIGHTLY_UNSAFE:
			pix = gdk_pixbuf_new_from_xpm_data (green_xpm);
			break;
		case GRG_UNSAFE:
		default:
			pix = gdk_pixbuf_new_from_xpm_data (red_xpm);
			break;
	}

	img = gtk_image_new_from_pixbuf (pix);
	gtk_misc_set_alignment (GTK_MISC(img), 0.5f, 0.0f);
	g_object_unref (G_OBJECT(pix));

	return img;
}

gchar
*grg_get_security_text (gchar* pattern)
{
	switch (safety_level){
		case GRG_SAFE:
			return g_strdup_printf(pattern, _("optimal"));
		case GRG_SLIGHTLY_UNSAFE:
			return g_strdup_printf(pattern, _("good"));
		case GRG_UNSAFE:
		default:
			return g_strdup_printf(pattern, _("low"));
	}
}

#define ADD_INDICATOR(box, text, pixbuf) \
	{ \
		GtkWidget *new_lbl, *img, *hbox; \
		new_lbl = gtk_label_new(text); \
		gtk_misc_set_alignment(GTK_MISC(new_lbl), 0, 0.5); \
		img = gtk_image_new_from_pixbuf (pixbuf); \
		hbox = gtk_hbox_new (FALSE, GRG_PAD); \
		gtk_box_pack_start(GTK_BOX(hbox), img, FALSE, FALSE, GRG_PAD); \
		gtk_box_pack_start(GTK_BOX(hbox), new_lbl, FALSE, FALSE, GRG_PAD); \
		gtk_box_pack_start(GTK_BOX(box), hbox, FALSE, FALSE, GRG_PAD); \
	}

void
grg_security_monitor (void)
{
	GtkWidget *dialog = gtk_dialog_new_with_buttons (_("Security monitor"),
					      NULL,
					      GTK_DIALOG_MODAL |
					      GTK_DIALOG_DESTROY_WITH_PARENT,
						  GTK_STOCK_OK,
					      GTK_RESPONSE_OK,
					      NULL);
	GdkPixbuf *red = gdk_pixbuf_new_from_xpm_data (red_xpm);
	GdkPixbuf *yellow = gdk_pixbuf_new_from_xpm_data (yellow_xpm);
	GdkPixbuf *green = gdk_pixbuf_new_from_xpm_data (green_xpm);
	struct rlimit *rl = (struct rlimit *) malloc(sizeof(struct rlimit));
	
	if (geteuid () && getegid () && getuid () && getgid ())
		ADD_INDICATOR (GTK_DIALOG (dialog)->vbox, _("Running without root privileges"), green)
	else
		ADD_INDICATOR (GTK_DIALOG (dialog)->vbox, _("Running without root privileges"), red)

	getrlimit(RLIMIT_CORE, rl);
	if(rl->rlim_cur || rl->rlim_max)
		ADD_INDICATOR (GTK_DIALOG (dialog)->vbox, _("Memory protection from core dumps"), red)
	else
		ADD_INDICATOR (GTK_DIALOG (dialog)->vbox, _("Memory protection from core dumps"), green)
	g_free(rl);

#ifdef HAVE_MLOCK
	if(mem_safe)
		ADD_INDICATOR (GTK_DIALOG (dialog)->vbox, _("Memory protection from swap writings"), green)
	else
		ADD_INDICATOR (GTK_DIALOG (dialog)->vbox, _("Memory protection from swap writings"), yellow) //the pwd isn't written in clear anyway
#endif

	if(ptrace_safe)
		ADD_INDICATOR (GTK_DIALOG (dialog)->vbox, _("Memory protection from ptrace spying"), green)
	else
		ADD_INDICATOR (GTK_DIALOG (dialog)->vbox, _("Memory protection from ptrace spying"), red)

#if defined(ENV_CHECK) && (defined(HAVE_CLEARENV) || defined(HAVE_ENVIRON))
	ADD_INDICATOR (GTK_DIALOG (dialog)->vbox, _("Environmental variables validation"), green)
#else
	ADD_INDICATOR (GTK_DIALOG (dialog)->vbox, _("Environmental variables validation"), red)
#endif

	ADD_INDICATOR (GTK_DIALOG (dialog)->vbox, _("stdout/stdin/stderr validation"), green)

#ifdef STRONG_RANDOM
	ADD_INDICATOR (GTK_DIALOG (dialog)->vbox, _("Enforced use of /dev/random"), green)
#else
	ADD_INDICATOR (GTK_DIALOG (dialog)->vbox, _("Enforced use of /dev/random"), yellow)
#endif

#ifdef ROOT_FILTER
	ADD_INDICATOR (GTK_DIALOG (dialog)->vbox, _("Strict prohibition to root user"), green)
#else
	ADD_INDICATOR (GTK_DIALOG (dialog)->vbox, _("Strict prohibition to root user"), yellow)
#endif

	gtk_widget_show_all (dialog);
	gtk_dialog_run (GTK_DIALOG (dialog));
	g_object_unref (G_OBJECT(red));
	g_object_unref (G_OBJECT(yellow));
	g_object_unref (G_OBJECT(green));
	gtk_widget_destroy (dialog);
}

/* XPM */
const gchar *red_xpm[] = {
"20 24 33 1",
"       c None",
".      c #1F1313",
"+      c #7F0A0A",
"@      c #ACA6A6",
"#      c #7A7A7A",
"$      c #CF0404",
"%      c #9D7676",
"&      c #461010",
"*      c #7E5E5E",
"=      c #DE7070",
"-      c #A00808",
";      c #675F5F",
">      c #AF5A5A",
",      c #630E0E",
"'      c #D65656",
")      c #E1A2A2",
"!      c #BB0606",
"~      c #F10202",
"{      c #472E2E",
"]      c #363232",
"^      c #8F0A0A",
"/      c #E10202",
"(      c #560E0E",
"_      c #AA2626",
":      c #D02222",
"<      c #6A2A2A",
"[      c #301212",
"}      c #710D0D",
"|      c #AE0606",
"1      c #862A2A",
"2      c #962A2A",
"3      c #BA2626",
"4      c #C3A6A6",
"                    ",
"                    ",
"      4>1+^2>)      ",
"    4<,}+^--||3)    ",
"   %&,}}+^-||!!$=   ",
"  #&(,}+^-||!$$$$=  ",
" @&((}++^-|!$$///$) ",
" {&&,,+^^|!!$//~~/: ",
"@[&&,}}^--!$$/~~~//)",
";[&((}+^^||$$/~~~/$'",
"][&&(,++--|!$//~//$:",
".[[((,++^-|!!$/$$$!!",
".[[&(,}+^--|!!$$$!!|",
"][[&((,++^-|||!!!!|_",
";.[&&(,}++^--|-|-|->",
"@..[&((,}++^^^----^)",
" ].[[&((,}++^+^^^^2 ",
" @.[[[&((,,}++++++4 ",
"  #..[&&((,,,}}}}%  ",
"   *.[[[&&((((,,%   ",
"    @][[[&&&((<4    ",
"      @;{[&{*@      ",
"                    ",
"                    "};

const gchar *yellow_xpm[] = {
"20 24 33 1",
"       c None",
".      c #1F1E18",
"+      c #A1984B",
"@      c #514D2C",
"#      c #716B36",
"$      c #E0D45E",
"%      c #CABE55",
"&      c #8D863F",
"*      c #ACACA9",
"=      c #2C2B1C",
"-      c #E4E1BD",
";      c #BFB452",
">      c #C5C3B3",
",      c #3B3825",
"'      c #5D582E",
")      c #7E783B",
"!      c #ABA24A",
"~      c #DCD6AA",
"{      c #7A7A7A",
"]      c #F3E463",
"^      c #474427",
"/      c #BEBAAE",
"(      c #929082",
"_      c #A8A688",
":      c #B7AC4E",
"<      c #6B6634",
"[      c #898240",
"}      c #DAD69A",
"|      c #D5C958",
"1      c #BAB682",
"2      c #6A6A64",
"3      c #978F42",
"4      c #656034",
"                    ",
"                    ",
"      >(&)&+1~      ",
"    /#4#)&3+!!;-    ",
"   (@4#)[&+!::;;}   ",
"  (@@4#)&++:;%%|%~  ",
" *,@'4#[&!!:%|$$$|- ",
" @,^'#)[3+:%|$$]$$| ",
"*=^@'<)[3!:;|$]]]$|-",
"2=,@'4)[3+:%|$]]]$|}",
",=^^'4)[3+!;||$]$$%%",
"==,@'<#)&+!:;||||%;:",
".=,^@4#)&3!!:;%%%%;!",
"=.=^@'4)[&3!!::;:!!:",
"2.=,^''<)[&3+!!!!!+1",
"_.==^@'4<)[&333+33&~",
" =.=,^@'4<#)[[[&&&3 ",
" *..=,^@'4<##)))))> ",
"  {.==,^@''4<###<_  ",
"   ).==,^^@'''4'_   ",
"    _,.=,,^^^@</    ",
"      *<^=,@)/      ",
"                    ",
"                    "};

const gchar *green_xpm[] = {
"20 24 33 1",
"       c None",
".      c #131F13",
"+      c #0A7A0A",
"@      c #A6ACA6",
"#      c #07CC07",
"$      c #104310",
"%      c #A2E1A2",
"&      c #06A606",
"*      c #7A7A7A",
"=      c #105010",
"-      c #76A876",
";      c #769276",
">      c #70DE70",
",      c #0B8E0B",
"'      c #0F5F0F",
")      c #5F675F",
"!      c #02F102",
"~      c #5ABA5A",
"{      c #0E700E",
"]      c #56D656",
"^      c #08BD08",
"/      c #5A9A5A",
"(      c #02E102",
"_      c #5E7E5E",
":      c #122612",
"<      c #A6C3A6",
"[      c #0C9A0C",
"}      c #323632",
"|      c #05D705",
"1      c #0B850B",
"2      c #2E432E",
"3      c #123312",
"4      c #08B108",
"                    ",
"                    ",
"      </11,[~%      ",
"    <{'{1,,&&&^%    ",
"   ;='{+1[[44^^#>   ",
"  ;=''{1,[&4^####>  ",
" @3='{+1[&4^#||((|% ",
" 2$=''+1[&4##|!!((| ",
"@3$='{+,[&4#|(!!!(|%",
"):$=={+1[&^^|(!!!(|]",
"233='{+1[&4^#|(!(|##",
"::$=='{1,[&^#||||##^",
".:3$''{1,[&44^###^44",
"}:3$=''+1,[&44^4^444",
").33=='{+1,[&&&4&&&~",
"@.:3$=''{+1,,[[[[[[<",
" }:33$=''{+11,,,,,[ ",
" @.:3$==''{{+++1++< ",
"  *.:33===''{{{{{-  ",
"   ).:33$===''''-   ",
"    @}:33$$==={<    ",
"      @)23$2_@      ",
"                    ",
"                    "};


const gchar *optimal_xpm[] = {
"20 24 69 1",
" 	c None",
".	c #A6C3A6",
"+	c #5A9A5A",
"@	c #0B850B",
"#	c #0B8E0B",
"$	c #0C9A0C",
"%	c #5ABA5A",
"&	c #A2E1A2",
"*	c #0E700E",
"=	c #0F5F0F",
"-	c #06A606",
";	c #08BD08",
">	c #769276",
",	c #105010",
"'	c #0A7A0A",
")	c #08B108",
"!	c #07CC07",
"~	c #70DE70",
"{	c #3D9F05",
"]	c #AC4402",
"^	c #A6ACA6",
"/	c #123312",
"(	c #05D705",
"_	c #8F6002",
":	c #AA4B00",
"<	c #1DC801",
"[	c #2E432E",
"}	c #104310",
"|	c #22B506",
"1	c #AB4701",
"2	c #AA5000",
"3	c #02F102",
"4	c #02E102",
"5	c #E21800",
"6	c #E21900",
"7	c #1DD601",
"8	c #5F675F",
"9	c #122612",
"0	c #23A807",
"a	c #E21500",
"b	c #56D656",
"c	c #AC3B02",
"d	c #FF0000",
"e	c #598804",
"f	c #2A470E",
"g	c #434A0B",
"h	c #AE3304",
"i	c #905403",
"j	c #131F13",
"k	c #5F2C0A",
"l	c #E30A01",
"m	c #5E4A09",
"n	c #5C5E07",
"o	c #C72501",
"p	c #323632",
"q	c #AF1A05",
"r	c #246C08",
"s	c #E30F01",
"t	c #26890A",
"u	c #E30901",
"v	c #E30C01",
"w	c #913B04",
"x	c #792C08",
"y	c #5F3F0A",
"z	c #942307",
"A	c #443E0C",
"B	c #7A7A7A",
"C	c #76A876",
"D	c #5E7E5E",
"                    ",
"                    ",
"      .+@@#$%&      ",
"    .*=*@##---;&    ",
"   >,=*'@$$));;!~   ",
"  >,==*@#$-);!!{]~  ",
" ^/,=*'@$-);!(_:<(& ",
" [},=='@$-)!|12344( ",
"^/},=*'#$-)|567334(&",
"89},,*'@$-0a5<3334(b",
"[//,=*'@$-cde(434(!!",
"99},fg*@#hdi!((((!!;",
"j9/kldm@ndo));!!!;))",
"p9/}qdlrsdt-)););)))",
"8j//,udvdw#$---)---%",
"^j9/}xddvr@##$$$$$$.",
" p9//}ddy*'@@#####$ ",
" ^j9/}zA==**'''@''. ",
"  Bj9//,,,==*****C  ",
"   8j9//},,,====C   ",
"    ^p9//}},,,*.    ",
"      ^8[/}[D^      ",
"                    ",
"                    "};
