/*  Gringotts - a small utility to safe-keep sensitive data
 *  (c) 2002, Germano Rizzo <mano@pluto.linux.it>
 *
 *  grg_widgets.c - various "custom" widgets used in Gringotts
 *  Author: Germano Rizzo
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
#include <string.h>

#include <gtk/gtk.h>

#include "grg_defs.h"

/**
 * grg_msg:
 * @text: the text of the message to display
 * @msgt: the GtkMessageType of the message
 * @parent: the parent of the dialog to create
 *
 * Displays a dialog that shows a message.
 */
void
grg_msg (gchar * text, GtkMessageType msgt, GtkWidget * parent)
{
	GtkWidget *dialog = gtk_message_dialog_new (GTK_WINDOW (parent), GTK_DIALOG_MODAL, msgt, GTK_BUTTONS_OK, text);

	gtk_dialog_set_default_response (GTK_DIALOG (dialog),
					 GTK_RESPONSE_OK);
	gtk_dialog_run (GTK_DIALOG (dialog));

	gtk_widget_destroy (dialog);
}

/**
 * grg_input_dialog:
 * @title: the title of the input dialog
 * @qtext: the question to ask
 * @preset: the predefined answer, NULL if none
 * @is_pass: tells if the requested string is a password, to be obfuscated
 * @parent: the parent of the dialog to create
 *
 * Shows and manages a dialog that asks for a string.
 *
 * Returns: the answer string, or NULL if user pressed Cancel.
 */
gchar *
grg_input_dialog (gchar * title, gchar * qtext, gchar * preset,
		  gboolean is_pass, GtkWidget * parent)
{
	GtkWidget *dialog, *question, *label;
	gchar *ret = NULL;
	gint res;

	dialog = gtk_dialog_new_with_buttons (title, GTK_WINDOW (parent),
					      GTK_DIALOG_MODAL, GTK_STOCK_OK,
					      GTK_RESPONSE_OK,
					      GTK_STOCK_CANCEL,
					      GTK_RESPONSE_CANCEL, NULL);

	gtk_dialog_set_default_response (GTK_DIALOG (dialog),
					 GTK_RESPONSE_OK);

	gtk_container_set_border_width (GTK_CONTAINER (dialog), GRG_PAD);
	gtk_box_set_spacing (GTK_BOX (GTK_DIALOG (dialog)->vbox), GRG_PAD);

	label = gtk_label_new (qtext);

	question = gtk_entry_new ();
	gtk_entry_set_max_length (GTK_ENTRY (question), 32);
	if (preset != NULL)
		gtk_entry_set_text (GTK_ENTRY (question), preset);
	gtk_entry_set_visibility (GTK_ENTRY (question), !is_pass);

	gtk_box_pack_start_defaults (GTK_BOX (GTK_DIALOG (dialog)->vbox),
				     label);
	gtk_box_pack_start_defaults (GTK_BOX (GTK_DIALOG (dialog)->vbox),
				     question);
	
	gtk_widget_grab_focus (question);
	gtk_widget_show_all (dialog);

	res = gtk_dialog_run (GTK_DIALOG (dialog));

	if (res == GTK_RESPONSE_OK)
		ret = g_strdup ((gchar *)
				gtk_entry_get_text (GTK_ENTRY (question)));

	gtk_widget_destroy (dialog);

	return ret;
}

/**
 * grg_ask_dialog:
 * @title: the title of the dialog
 * @question: the question to ask
 * @allowcanc: if the dialog allows a cancel button or not
 * @parent: the parent of the dialog to create.
 *
 * Asks a question for a boolean answer.
 *
 * Returns: GRG_YES if yes, GRG_NO if no, GRG_CANCEL if user canceled
 */
grg_response
grg_ask_dialog (gchar * title, gchar * question, gboolean allowcanc,
		GtkWidget * parent)
{
	GtkWidget *dialog, *label;
	gint res;

	if (!allowcanc)
		dialog = gtk_dialog_new_with_buttons (title,
						      GTK_WINDOW (parent),
						      GTK_DIALOG_MODAL,
						      GTK_STOCK_YES,
						      GTK_RESPONSE_YES,
						      GTK_STOCK_NO,
						      GTK_RESPONSE_NO, NULL);
	else
		dialog = gtk_dialog_new_with_buttons (title,
						      GTK_WINDOW (parent),
						      GTK_DIALOG_MODAL,
						      GTK_STOCK_YES,
						      GTK_RESPONSE_YES,
						      GTK_STOCK_NO,
						      GTK_RESPONSE_NO,
						      GTK_STOCK_CANCEL,
						      GTK_RESPONSE_CANCEL,
						      NULL);

	gtk_dialog_set_default_response (GTK_DIALOG (dialog),
					 GTK_RESPONSE_OK);

	gtk_container_set_border_width (GTK_CONTAINER (dialog), GRG_PAD);
	gtk_box_set_spacing (GTK_BOX (GTK_DIALOG (dialog)->vbox), GRG_PAD);

	label = gtk_label_new (question);

	gtk_box_pack_start_defaults (GTK_BOX (GTK_DIALOG (dialog)->vbox),
				     label);

	gtk_widget_show_all (dialog);

	res = gtk_dialog_run (GTK_DIALOG (dialog));

	gtk_widget_destroy (dialog);

	switch (res)
	{
	case GTK_RESPONSE_YES:
		return GRG_YES;
	case GTK_RESPONSE_NO:
		return GRG_NO;
	case GTK_RESPONSE_CANCEL:
	default:
		return GRG_CANCEL;
	}
}

/**
 * grg_window_set_icon:
 * @w: the window to set the icon to
 *
 * Associates the Gringotts icon with the given window.
 */
void
grg_window_set_icon (GtkWindow * w)
{
	GdkPixbuf *gp = gdk_pixbuf_new_from_file (PIX_DIR "gringotts.xpm", NULL);
	if (gp)
	{
		GList *gl = NULL;

		g_list_append (gl, gp);
		gtk_window_set_icon (w, gp);
		gtk_window_set_default_icon_list (gl);

		g_list_free (gl);
		g_object_unref (G_OBJECT (gp));
	}
	else
		g_warning (_("Icon file %s not found"), PIX_DIR "gringotts.xpm");
}

/**
 * grg_find_dialog:
 */
gboolean
grg_find_dialog (guchar ** needle, gboolean * only_current,
		 gboolean * case_sens, GtkWindow * parent)
{
	GtkWidget *dialog, *question, *label, *chk1, *chk2;
	gint res;

	dialog = gtk_dialog_new_with_buttons (_("Find a string"),
					      GTK_WINDOW (parent),
					      GTK_DIALOG_MODAL, GTK_STOCK_OK,
					      GTK_RESPONSE_OK,
					      GTK_STOCK_CANCEL,
					      GTK_RESPONSE_CANCEL, NULL);

	gtk_dialog_set_default_response (GTK_DIALOG (dialog),
					 GTK_RESPONSE_OK);

	gtk_container_set_border_width (GTK_CONTAINER (dialog), 3);
	gtk_box_set_spacing (GTK_BOX (GTK_DIALOG (dialog)->vbox), 3);

	label = gtk_label_new (_("Enter the string to find:"));

	question = gtk_entry_new ();
	gtk_entry_set_max_length (GTK_ENTRY (question), 32);
	if (*needle)
		gtk_entry_set_text (GTK_ENTRY (question), *needle);

	chk1 = gtk_check_button_new_with_label (_
						("Search only in this entry"));
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (chk1),
				      *only_current);

	chk2 = gtk_check_button_new_with_label (_("Case sensitive"));
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (chk2), *case_sens);

	gtk_box_pack_start_defaults (GTK_BOX (GTK_DIALOG (dialog)->vbox),
				     label);
	gtk_box_pack_start_defaults (GTK_BOX (GTK_DIALOG (dialog)->vbox),
				     question);
	gtk_box_pack_start_defaults (GTK_BOX (GTK_DIALOG (dialog)->vbox),
				     chk1);
	gtk_box_pack_start_defaults (GTK_BOX (GTK_DIALOG (dialog)->vbox),
				     chk2);

	gtk_widget_grab_focus (question);
	gtk_widget_show_all (dialog);

	res = gtk_dialog_run (GTK_DIALOG (dialog));

	if (res != GTK_RESPONSE_OK)
	{
		gtk_widget_destroy (dialog);
		return FALSE;
	}

	g_free (*needle);
	*needle = g_strdup (gtk_entry_get_text (GTK_ENTRY (question)));
	*only_current =
		gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (chk1));
	*case_sens = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (chk2));

	gtk_widget_destroy (dialog);

	return TRUE;
}

#define ADD_LIB_VERSION(box, text, v1, v2, v3) \
	{ \
		GtkWidget *new_lbl; \
		gchar *lbl = g_strdup_printf("%s: %d.%d.%d", text, v1, v2, v3); \
		new_lbl = gtk_label_new(lbl); \
		gtk_misc_set_alignment(GTK_MISC(new_lbl), 0, 0.5); \
		gtk_box_pack_start(GTK_BOX(box), new_lbl, FALSE, FALSE, GRG_PAD); \
		g_free (lbl); \
	}
